/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use std::sync::Arc;

use color_eyre::eyre::{Result, WrapErr};
use crossterm::event::{MouseEvent, MouseEventKind};
use ratatui::{prelude::*, widgets::*};
use tokio::sync::mpsc::UnboundedSender;
use tokio::sync::OnceCell;

use super::Component;
use crate::{
  action::Action,
  buffers::{Buffer, Buffers, FullyReadStatus},
  config::{Config, KeyBindings},
};

pub struct Buflist {
  command_tx: Option<UnboundedSender<Action>>,
  config: Arc<Config>,

  /// Updated by [`draw`], used by [`handle_mouse_events`] to find which buffer was clicked.
  last_area: Option<Rect>,
  last_num_buffers: Option<usize>,
}
impl Buflist {
  pub fn new(config: Arc<Config>) -> Self {
    Buflist {
      command_tx: None,
      config,
      last_area: None,
      last_num_buffers: None,
    }
  }

  pub fn num_columns(&mut self, area: Rect, buffers: &Buffers) -> u16 {
    let borders_height = 2;
    buffers
      .len()
      .div_ceil((area.height - borders_height).into())
      .try_into()
      .unwrap_or(u16::MAX)
  }

  pub fn borders_width(&self) -> u16 {
    2
  }

  pub fn get_room_name_style(&self, buf: &dyn Buffer) -> Style {
    use matrix_sdk::sync::UnreadNotificationsCount;
    let config = &self.config.style.buflist;
    match buf.unread_notification_counts() {
      UnreadNotificationsCount {
        highlight_count: 0,
        notification_count: 0,
      } => match buf.fully_read() {
        FullyReadStatus::Not => *config.unread_message,
        FullyReadStatus::OnlyMessages => *config.unread_event,
        FullyReadStatus::All => *config.uneventful,
      },
      UnreadNotificationsCount {
        highlight_count: 0,
        notification_count: _,
      } => *config.notification,
      UnreadNotificationsCount {
        highlight_count: _,
        notification_count: _,
      } => *config.highlight,
    }
  }
}

impl Component for Buflist {
  fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
    self.command_tx = Some(tx);
    Ok(())
  }

  fn init(&mut self, area: Rect) -> Result<()> {
    self.last_area = Some(area);
    Ok(())
  }

  fn handle_mouse_events(&mut self, mouse: MouseEvent) -> Result<Option<Action>> {
    match mouse {
      MouseEvent {
        kind: MouseEventKind::Up(_),
        column: x,
        row: y,
        modifiers: _,
      } => {
        let Some(last_area) = self.last_area else {
          return Ok(None);
        };
        if !last_area.intersects(Rect {
          x,
          y,
          height: 1,
          width: 1,
        }) {
          // Clicked outside the buflist (or on the border)
          return Ok(None);
        }
        let Some(last_num_buffers) = self.last_num_buffers else {
          return Ok(None);
        };
        let column = (x - last_area.x) / self.config.layout.buflist.column_width;
        let mut buffer_index: isize =
          (column as isize) * (last_area.height as isize) + ((y as isize) - (last_area.y as isize));
        let buffer_index_usize: usize = buffer_index.try_into().expect("negative buffer_index");
        if buffer_index_usize >= last_num_buffers {
          if self.config.layout.buflist.penultimate_right_overflow {
            // Clicked on the overflow of the last-but-one column
            buffer_index -= last_area.height as isize;
          } else {
            // Clicked past the last buffer
            return Ok(None);
          }
        }
        let buffer_index_usize: usize = buffer_index.try_into().expect("negative buffer_index");
        if buffer_index_usize >= last_num_buffers {
          tracing::error!("[BUG] Unexpected click past the last buffer");
          return Ok(None);
        }
        Ok(Some(Action::ToBuffer(buffer_index)))
      },
      _ => Ok(None),
    }
  }

  fn update(&mut self, action: &Action) -> Result<Option<Action>> {
    Ok(None)
  }

  fn draw(
    &mut self,
    frame: &mut Frame<'_>,
    area: Rect,
    buffers: &crate::buffers::Buffers,
  ) -> Result<()> {
    let block = Block::new().borders(Borders::ALL);
    let inner_area = block.inner(area);
    block.render(area, frame.buffer_mut());

    let area = inner_area;
    self.last_area = Some(area);
    self.last_num_buffers = Some(buffers.len());

    let num_digits = u32::min(5, buffers.len().ilog10() + 1) as usize;
    let right_pad = " ".repeat(area.width.into());
    let mut stack = Vec::new(); // List of parent buffers of the current one
    let mut buffers_iter = buffers.iter().enumerate();
    let max_columns = u16::min(
      self.config.layout.buflist.max_columns,
      area.width.div_ceil(self.config.layout.buflist.column_width),
    );
    let column_width = self.config.layout.buflist.column_width;
    for col in 0..max_columns {
      let x = area.y + col * column_width;
      let allow_overflow = if col == max_columns - 2 {
        self.config.layout.buflist.penultimate_right_overflow
      } else if col == max_columns - 1 {
        // Always allow the last column to overflow
        true
      } else {
        false
      };
      for y in area.top()..area.bottom() {
        let Some((i, buf)) = buffers_iter.next() else {
          return Ok(());
        };
        match buffers
          .parents()
          .get(&buf.id())
          .into_iter()
          .flatten()
          .flat_map(|parent| stack.iter().position(|id| *id == *parent))
          .max() // if in multiple spaces, the last one is most likely to matter
        {
          Some(parent_position) => stack.truncate(parent_position + 1),
          None => stack.clear(),
        }
        stack.push(buf.id());
        let buf_number = format!("{}.", i + 1);
        let mut base_style = Style::default();
        let mut name_style = self.get_room_name_style(buf);
        if i == buffers.active_index() {
          base_style = base_style.on_blue();
          name_style = name_style.on_blue();
        } else {
          base_style = base_style.bg(Color::Reset);
          name_style = name_style.bg(Color::Reset);
        }
        let tree_pad = " ".repeat(stack.len() - 1);
        let buf_number_style = base_style.green();
        let left_pad = " ".repeat((num_digits + 1).saturating_sub(buf_number.len()));
        frame.render_widget(
          Paragraph::new::<Line<'_>>(
            vec![
              Span::styled(left_pad, base_style),
              Span::styled(buf_number, buf_number_style),
              Span::styled(tree_pad, base_style),
              Span::styled(buf.short_name(), name_style),
              Span::styled(right_pad.clone(), base_style),
            ]
            .into(),
          ),
          Rect {
            x,
            y,
            width: if allow_overflow && y as usize > buffers.len() % (area.height as usize) {
              area.width
            } else {
              column_width
            },
            height: 2,
          }
          .intersection(area),
        );
      }
    }

    // TODO: Do something with the remaining buffers, if any. eg. print "XX more buffers"
    // at the end or display a scrollbar

    Ok(())
  }
}
