/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use std::borrow::Cow;
use std::hash::{Hash, Hasher};
use std::rc::Rc;

use html5ever::driver::parse_fragment;
use html5ever::interface::QualName;
use html5ever::tendril::TendrilSink;
use html5ever::{local_name, namespace_url, ns, Attribute};
use markup5ever_rcdom::{Handle, Node, NodeData, RcDom};
use ratatui::style::{Color, Style, Stylize};
use ratatui::text::{Line, Span, Text};

use crate::config::Config;

pub fn escape_html<S: ?Sized + AsRef<str>>(s: &S) -> Cow<'_, str> {
  html_escape::encode_text_minimal(s)
}

pub fn markup_colored_by_mxid<M: ?Sized + AsRef<str>, C: ?Sized + AsRef<str>>(
  mxid: &M,
  content: &C,
) -> String {
  format!(
    r#"<font data-ratatrix-colored-by-mxid="{}">{}</font>"#,
    html_escape::encode_double_quoted_attribute(mxid),
    content.as_ref()
  )
}

fn count_digits(n: u16) -> u8 {
  f32::log10(n.into()).floor() as u8 + 1
}

fn get_color_by_mxid(config: &Config, mxid: &str) -> Option<Style> {
  let rotation = &config.style.users.rotation;
  if rotation.is_empty() {
    None
  } else {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();
    mxid.hash(&mut hasher);
    let color_id: usize = (hasher.finish() & (usize::MAX as u64)) as usize;
    Some(*rotation[color_id % rotation.len()])
  }
}

fn parse_hex_color(s: &str) -> Option<Color> {
  // Should we implement a workaround for https://github.com/ratatui-org/ratatui/issues/475
  // here, or wait for Crossterm to fix it?
  let s = s.strip_prefix('#')?;
  let a: [u8; 3] = hex::FromHex::from_hex(s).ok()?;
  let a = [0, a[0], a[1], a[2]];
  let n = u32::from_be_bytes(a);
  Some(Color::Rgb(
    (n >> 16) as u8,
    ((n & 0xFF00) >> 8) as u8,
    (n & 0xFF) as u8,
  ))
}

#[derive(Clone, Debug, Default)]
enum ListState {
  #[default]
  None,
  Unordered,
  Ordered {
    counter: u16,
    digits: u8,
  },
}

#[derive(Clone, Debug, Default)]
struct FormatState {
  style: Style,
  padding: String,
  list_state: ListState,
}

fn format_tree(
  config: &Config,
  tree: Rc<Node>,
  state: &mut FormatState,
  text: &mut Vec<(String, Text<'static>)>,
  mut previous_sibling_is_block: bool,
) -> bool {
  let mut state = match &tree.data {
    NodeData::Document
    | NodeData::Doctype { .. }
    | NodeData::Comment { .. }
    | NodeData::ProcessingInstruction { .. } => state.to_owned(),
    NodeData::Text { contents } => {
      let s: String = contents.clone().into_inner().into();
      let s = s.replace('\n', ""); // Lines are insignificant in HTML
      if previous_sibling_is_block && !s.is_empty() {
        text.push((state.padding.clone(), Text::raw("")));
        previous_sibling_is_block = false;
      }
      text
        .last_mut()
        .unwrap()
        .1
        .lines
        .last_mut()
        .unwrap()
        .spans
        .push(Span::styled(s, state.style));
      state.to_owned()
    },
    NodeData::Element {
      name: QualName {
        ns: ns!(html),
        local: name,
        ..
      },
      attrs,
      ..
    } => match *name {
      local_name!("font") | local_name!("span") => {
        let mut state = state.to_owned();
        for attr in attrs.borrow().iter() {
          match attr {
            Attribute {
              name:
                QualName {
                  ns: ns!(html),
                  local: name,
                  ..
                },
              value,
            }
            | Attribute {
              name:
                QualName {
                  ns: ns!(),
                  local: name,
                  ..
                },
              value,
            } => match name.as_ref() {
              "data-mx-color" => {
                if let Some(color) = parse_hex_color(value.as_ref()) {
                  state.style.fg = Some(color);
                }
              },
              "data-mx-bg-color" => {
                if let Some(color) = parse_hex_color(value.as_ref()) {
                  state.style.bg = Some(color);
                }
              },
              "data-ratatrix-colored-by-mxid" => {
                if let Some(style) = get_color_by_mxid(config, value.as_ref()) {
                  state.style = state.style.patch(style);
                }
              },
              _ => {},
            },
            _ => {},
          }
        }
        state
      },
      local_name!("br") => {
        text.push((state.padding.clone(), Text::raw("")));
        state.to_owned()
      },
      local_name!("p") => {
        previous_sibling_is_block = true;
        state.to_owned()
      },
      local_name!("blockquote") => {
        previous_sibling_is_block = true;
        FormatState {
          padding: state.padding.to_owned() + "> ",
          ..state.to_owned()
        }
      },
      local_name!("ul") => {
        previous_sibling_is_block = true;
        FormatState {
          list_state: ListState::Unordered,
          ..state.to_owned()
        }
      },
      local_name!("ol") => {
        previous_sibling_is_block = true;
        // FIXME: <li> are not guaranteed to be direct children, are they?
        let items_count: u16 = tree
          .children
          .borrow()
          .iter()
          .map(|child| match &child.data {
            NodeData::Element {
              name:
                QualName {
                  ns: ns!(html),
                  local: local_name!("li"),
                  ..
                },
              ..
            } => 1,
            _ => 0,
          })
          .sum();
        FormatState {
          list_state: ListState::Ordered {
            counter: 0,
            digits: count_digits(items_count),
          },
          ..state.to_owned()
        }
      },
      local_name!("li") => {
        previous_sibling_is_block = false;
        match state.list_state {
          ListState::None => state.to_owned(),
          ListState::Unordered => {
            let mut line = Line::default();
            line.spans.push(Span::styled("* ", state.style));
            text.push((state.padding.to_owned(), line.into()));
            FormatState {
              padding: state.padding.to_owned() + "  ",
              ..state.to_owned()
            }
          },
          ListState::Ordered {
            ref mut counter,
            digits,
          } => {
            let mut line = Line::default();
            *counter += 1;
            line
              .spans
              .push(Span::styled(format!("{}. ", counter), state.style));
            if count_digits(*counter) != digits {
              line.spans.push(Span::styled(
                " ".repeat((digits - count_digits(*counter)).into()),
                state.style,
              ));
            }
            text.push((state.padding.to_owned(), line.into()));
            FormatState {
              padding: state.padding.to_owned() + "   " + &" ".repeat(digits.into()),
              ..state.to_owned()
            }
          },
        }
      },
      local_name!("em") | local_name!("i") => FormatState {
        style: state.style.italic(),
        ..state.to_owned()
      },
      local_name!("strong") | local_name!("b") => FormatState {
        style: state.style.bold(),
        ..state.to_owned()
      },
      local_name!("u") => FormatState {
        style: state.style.underlined(),
        ..state.to_owned()
      },
      _ => state.to_owned(),
    },
    NodeData::Element { .. } => state.to_owned(), // Element not in the HTML namespace
  };

  for subtree in tree.children.borrow().iter() {
    previous_sibling_is_block = format_tree(
      config,
      subtree.clone(),
      &mut state,
      text,
      previous_sibling_is_block,
    );
  }

  match &tree.data {
    NodeData::Document
    | NodeData::Doctype { .. }
    | NodeData::Comment { .. }
    | NodeData::ProcessingInstruction { .. } => {},
    NodeData::Text { .. } => {},
    NodeData::Element {
      name: QualName {
        ns: ns!(html),
        local: name,
        ..
      },
      attrs,
      ..
    } => match name.as_ref() {
      "p" | "blockquote" => {
        previous_sibling_is_block = true;
      },
      _ => {},
    },
    NodeData::Element { .. } => {},
  }

  previous_sibling_is_block
}

/// Returns a list of pairs `(padding, text)`.
///
/// When rendering, the padding should be added left of every line of the text.
pub fn format_html(config: &Config, prefix: &'static str, s: &str) -> Vec<(String, Text<'static>)> {
  let tree = parse_fragment(
    RcDom::default(),
    Default::default(),
    QualName::new(None, ns!(html), local_name!("body")),
    Vec::new(),
  )
  .one(s)
  .document;
  let prefix = Text::raw(prefix);
  let mut state = FormatState {
    padding: " ".repeat(prefix.width() + 1), // TODO: make +1 configurable
    ..Default::default()
  };
  let mut text = vec![("".to_owned(), prefix)];
  format_tree(config, tree, &mut state, &mut text, false);
  text
}
